import sys
import pandas as pd
import numpy as np
import os
import json
from copy import deepcopy
import argparse
import datetime
import itertools
import tqdm

tree_template = {
    "path":"{path}",
    "count":1,
    "placement":{
        "method": "position",
        "x": "{x}",
        "y": "{y}"
    }
}

def get_tree_measurements(laz_file, geojson_dir):
    geojson_file = '_'.join(laz_file.split('_')[:3]) + '.geojson'
    geojson_path = os.path.join(geojson_dir, geojson_file)
    with open(geojson_path) as f:
        try:
            geojson = json.load(f)
            #print(*[m for m in geojson['properties']['measurements']], sep='\n')
            #dbh_cm = geojson['properties']['measurements'][0]['DBH_cm']
            arr = [m['height_m'] for m in geojson['properties']['measurements'] if 'height_m' in m and m['height_m'] is not None]
            height_m = np.nanmean(arr)
            arr = [m['DBH_cm'] for m in geojson['properties']['measurements'] if 'DBH_cm' in m and m['DBH_cm'] is not None]
            if len(arr) == 0:
                dbh_cm = np.nan
            else:
                dbh_cm = np.nanmean(arr)
            return dbh_cm, height_m
        except:
            print('Error reading geojson file: {}'.format(geojson_path))
            print(geojson['properties']['measurements'])

def get_all_tree_measurements(laz_files, geojson_dir):
    tree_measurements = pd.DataFrame(columns=['laz_file', 'dbh_cm', 'height_m'], index=range(len(laz_files)), dtype=float)
    for i, laz_file in enumerate(laz_files):
        dbh_cm, height_m = get_tree_measurements(laz_file, geojson_dir)
        tree_measurements.loc[i] = [laz_file, dbh_cm, height_m]
    tree_measurements.dropna(inplace=True)
    return tree_measurements

def find_closest_tree(d, h, measurements):
    measurements['dist'] = np.sqrt((measurements['height_m'] - h)**2 + (measurements['dbh_cm'] - d)**2)
    return measurements[measurements['dist'] == measurements['dist'].min()]

def scene(config):
    print("Building scene configuration files...")
    forest_stand_path = config['forest']
    pytreedb_dir = config['scene']['pytreedb_dir']
    geojson_dir = os.path.join(pytreedb_dir, 'geojsons')
    model_source = config['scene']['model_source']
    has_leaves = config['scene']['has_leaves']
    min_quality = config['scene']['min_quality']
    laz_dir = os.path.join(pytreedb_dir, model_source, 'laz')
    # Read the data from the file
    data = pd.read_csv(forest_stand_path, sep=',')
    # get indexes
    data = data.reset_index()
    n_trees = data.shape[0]
    print('Number of trees: {}'.format(n_trees))
    c = {'x': 'X', 'y': 'Y', 'h': 'H', 'd': 'D', 'spec': 'Species'}
    scene_template = config['scene']['template']
    #helios_template = json.loads(json.dumps(helios_template).replace("{timestamp}", config['timestamp']).replace("{id}", config['id']))
    total_x = data[c['x']].max() - data[c['x']].min()
    total_y = data[c['y']].max() - data[c['y']].min()
    scene_template["area"]["x"] = total_x
    scene_template["area"]["y"] = total_y

    df_trees = pd.DataFrame(columns=['x', 'y', 'd', 'h', 'spec', 'file_id', 'tile_id', 'tile_x', 'tile_y'], index=range(n_trees), dtype=float)

    tiled = False
    tiles_x = []
    tiles_y = []
    if "tile_w" in config["scene"]:
        tiled = True
        start_x = data[c['x']].min() - data[c['x']].min() % config['scene']['tile_w']
        start_y = data[c['y']].min() - data[c['y']].min() % config['scene']['tile_h']
        tiles_x = np.arange(start_x, data[c['x']].max(), config['scene']['tile_w'])
        tiles_y = np.arange(start_y, data[c['y']].max(), config['scene']['tile_h'])
    
    for tile_id, (tile_x, tile_y) in tqdm.tqdm(enumerate(itertools.product(tiles_x, tiles_y))):
        min_x = tile_x - config["scene"]['tile_overlap']
        max_x = tile_x + config["scene"]['tile_w'] + config['scene']['tile_overlap']
        min_y = tile_y - config["scene"]['tile_overlap']
        max_y = tile_y + config["scene"]['tile_h'] + config['scene']['tile_overlap']
        tile_data = data[(data[c['x']] >= min_x) & (data[c['x']] < max_x) & (data[c['y']] >= min_y) & (data[c['y']] < max_y)]
        if tile_data.shape[0] == 0:
            continue
        tile_scene_template = deepcopy(scene_template)
        tile_scene_template = json.loads(json.dumps(tile_scene_template).replace("{tile_id}", str(tile_id + 1)))
        tile_scene_template["assets_dir"] = '/forestcare/assets'
        tile_scene_template['data_dir'] = '/forestcare/scene'
        tile_scene_template["area"]["x"] = config['scene']['tile_w']
        tile_scene_template["area"]["y"] = config['scene']['tile_h']
        tile_scene_template["area"]["x_offset"] = tile_x
        tile_scene_template["area"]["y_offset"] = tile_y
        leaves_arg = 'on' if has_leaves else 'off'
        # print("Looking for laz files with leaves={} and min_quality={}...".format(leaves_arg, min_quality))
        for spec, cnt in tile_data[c['spec']].value_counts().items():
            all_laz_files = [f for f in os.listdir(laz_dir) if f.endswith('.laz') and f.startswith(spec)]
            filtered_laz_files = [f for f in all_laz_files if f.split('.')[-2].split('-')[-1] == leaves_arg and f.split('_')[-2] in ['q1', 'q2', 'q3','q4','q5'][min_quality-1:]]
            #print("Found {} valid laz files out of {} for species {} ".format(len(filtered_laz_files), len(all_laz_files), spec))
            tree_measurements = get_all_tree_measurements(filtered_laz_files, geojson_dir)
            for i, tree in tile_data[tile_data[c['spec']] == spec].iterrows():
                original_tree_id = tree['index']
                tree = tree[1:]
                dbh_cm, height_m = tree.D * 100, tree.H
                closest_measurements = find_closest_tree(dbh_cm, height_m, tree_measurements)
                laz_file = closest_measurements["laz_file"].values[0]
                tree_h = closest_measurements["height_m"].values[0]
                tree_d = closest_measurements["dbh_cm"].values[0]
                # append [tree[0], tree[1], tree_h, tree_d, spec, tile_id, tile_x, tile_y] to df_trees
                if tree[0] >= tile_x and tree[0] < tile_x + config['scene']['tile_w'] and tree[1] >= tile_y and tree[1] < tile_y + config['scene']['tile_h']:
                    df_trees.loc[original_tree_id] = [tree[0], tree[1], (tree_d / 100.0), tree_h, spec, laz_file, tile_id, tile_x, tile_y]
                cur_tree = deepcopy(tree_template)
                cur_tree['part_id'] = int(original_tree_id)
                cur_tree['path'] = cur_tree['path'].format(path=laz_file)
                cur_tree['placement']['x'] = cur_tree['placement']['x'].format(x=tree[0] - tile_x)
                cur_tree['placement']['y'] = cur_tree['placement']['y'].format(y=tree[1] - tile_y)
                tile_scene_template['trees'].append(cur_tree)
        output_path = os.path.join(config["work_dir"], config["id"], "scene", config["id"] + '_' + str(tile_id+1) + '.json')
        if not os.path.exists(os.path.dirname(output_path)):
            os.makedirs(os.path.dirname(output_path))
        with open(output_path, 'w') as f:
            json.dump(tile_scene_template, f, indent=4)
    df_trees.to_csv(os.path.join(config["work_dir"], config["id"], 'forest_scene.csv'), index=False)