import laspy
import os
import numpy as np
import sklearn.cluster as cluster
import sklearn.metrics as metrics
import itertools
import tqdm
import json
import sys
from copy import deepcopy

def load_tile(tile):
    laz = laspy.read(tile)
    return laz

def write_laz(path, laz):
    laz.write(path)

def load_tile_config(helios_configs_path, forest_basename, tile_id):
    helios_config = os.path.join(helios_configs_path, forest_basename + '_' + str(tile_id) + '.json')
    with open(helios_config, 'r') as f:
        config = json.load(f)
        x_offset = float(config['area']['x_offset'])
        y_offset = float(config['area']['y_offset'])
        return config, x_offset, y_offset

def get_trees(tile_config):
    trees = []
    for tree in tile_config['trees']:
        x,y = tree['placement']['x'], tree['placement']['y']
        x,y = float(x), float(y)
        trees.append([x, y])
    return trees

def get_tile_mapping(unique_trees, tile_config, x_offset, y_offset):
    trees = get_trees(tile_config)
    # get tree mapping to unique trees
    tree_mapping = []
    for tree in trees:
        offset_tree = [tree[0] + x_offset, tree[1] + y_offset]
        tree_mapping.append(np.where(np.all(unique_trees == offset_tree, axis=1))[0][0])
    return np.array(tree_mapping)

def merge_tiles(forest_basename, tile_configs, out_dir, work_dir, lastools_dir, fix_edges=True):
    out_path = os.path.join(out_dir, forest_basename, forest_basename + '.laz')
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    if not fix_edges:
        out_tiles_dir = os.path.join(out_dir, forest_basename, 'tiles_with_edge')
    else:
        out_tiles_dir = os.path.join(out_dir, forest_basename, 'tiles')
    if not os.path.exists(out_tiles_dir):
        os.makedirs(out_tiles_dir)
    out_tile_path = os.path.join(out_tiles_dir, forest_basename + '_{tile_id}.laz')
    laz_tile_path = "{work_dir}/simulation/results/Survey Playback/{forest_basename}_{tile_id}/{child}/points/strip{forest_basename}_{tile_id}-full_points.las".replace("{forest_basename}", forest_basename).replace("{work_dir}", work_dir)
    tiles_dir = os.path.dirname(tile_configs[0])
    all_trees = []
    for tile_id in range(1, 1 + len(tile_configs)):
        tile_config = os.path.join(tiles_dir, forest_basename + '_' + str(tile_id) + '.json')
        with open(tile_config, 'r') as f:
            config = json.load(f)
            x_offset = float(config['area']['x_offset'])
            y_offset = float(config['area']['y_offset'])
            for tree in config['trees']:
                x,y = tree['placement']['x'], tree['placement']['y']
                x,y = float(x), float(y)
                # add to all_trees
                all_trees.append([x + x_offset, y + y_offset])

    # unique trees
    all_trees = np.array(all_trees)
    unique_trees = np.unique(all_trees, axis=0)

    for tile_id in tqdm.tqdm(range(1, 1 + len(tile_configs))):
        tile_file = laz_tile_path.format(tile_id=tile_id, child="*")
        tile_upper = tile_file.split("*")[0]
        tile_lower = tile_file.split("*")[1]
        tile_filepath = tile_upper + os.listdir(tile_upper)[0] + tile_lower
        tile_config, x_offset, y_offset = load_tile_config(tiles_dir, forest_basename, tile_id)
        tree_mapping = get_tile_mapping(unique_trees, tile_config, x_offset, y_offset)
        if not os.path.exists(tile_filepath):
            print("Tile file does not exist: {}".format(tile_file))
            break
        laz = load_tile(tile_filepath)
        #laz.hitObjectId = tree_mapping[laz.hitObjectId]
        # take only the points that are in the tile and not overlapping
        if fix_edges:
            x_start = float(tile_config['area']['x_offset'])
            x_end = x_start + float(tile_config['area']['x'])
            y_start = float(tile_config['area']['y_offset'])
            y_end = y_start + float(tile_config['area']['y'])
            x_mask = np.logical_and(laz.x >= x_start, laz.x <= x_end)
            y_mask = np.logical_and(laz.y >= y_start, laz.y <= y_end)
            mask = np.logical_and(x_mask, y_mask)
            laz = laz[mask]
        write_laz(out_tile_path.format(tile_id=tile_id), laz)

    # run las2las to merge tiles
    las2las = os.path.join(lastools_dir, 'bin', 'las2las')
    cmd = las2las + " -i " + out_tile_path.format(tile_id="*") + " -merged " + "-o " + out_path
    print(cmd)
    os.system(cmd)

if __name__ == '__main__':
    if len(sys.argv) < 4:
        print("Usage: python merge_tiles.py <work_dir> <out_dir> <lastools_dir>")
        sys.exit(1)
    work_dir = sys.argv[1]
    out_dir = sys.argv[2]
    lastools_dir = sys.argv[3]
    forest_basename = os.path.basename(work_dir)
    tiles_dir = '{work_dir}/scene'.format(work_dir=work_dir)
    tile_configs = [os.path.join(tiles_dir, tile) for tile in os.listdir(tiles_dir)]
    merge_tiles(forest_basename, tile_configs, out_dir, work_dir, lastools_dir, fix_edges=True)