import os
import sys
import json
import xml.etree.ElementTree as ET
import numpy as np
from copy import deepcopy
import pandas as pd
import laspy
import random

def help():
    print("USAGE: python build_scene.py scene.json simulation.json")

def get_tree_lazs(config):
    laz_dir = os.path.join(config["data_dir"], "laz")
    all_laz_files = [f for f in os.listdir(laz_dir) if f.split('.')[-1] == 'laz']
    laz_files = []
    for tree_set in config["scene"]["trees"]:
        species = tree_set["template"]
        # load laz file
        species_laz_files = [f for f in all_laz_files if f.startswith(species)]
        if len(species_laz_files) == 0:
            print("ERROR: laz file not found for species: %s" % species)
            sys.exit(1)
        # select one laz file randomly
        laz_file = os.path.join(laz_dir, random.choice(species_laz_files))
        #laz_file = os.path.join(laz_dir, species_laz_files[0])
        if not os.path.isfile(laz_file):
            print("ERROR: laz file not found: %s" % laz_file)
            sys.exit(1)
        laz_files.append(laz_file)
    xyz_files = []
    for i, laz_file in enumerate(laz_files):
        laz = laspy.read(laz_file)
        # save xyz as txt file
        xyz = np.vstack((laz.x, laz.y, laz.z)).transpose()
        # center x,y to 0,0 and make z min 0
        xyz = xyz - xyz.mean(axis=0)
        xyz[:,2] -= xyz[:,2].min()
        # save as .xyz file
        xyz_file = os.path.join(laz_dir, "%s_%03d.xyz" % (species, i))
        np.savetxt(xyz_file, xyz, fmt="%.6f")
        xyz_files.append(xyz_file)
    print("XYZ_Files: %s" % str(xyz_files))
    return xyz_files

def laz_to_xyz(laz_path, xyz_file = None, xyz_dir = None):
    laz = laspy.read(laz_path)
    # save xyz as txt file
    xyz = np.vstack((laz.x, laz.y, laz.z)).transpose()
    # center x,y to 0,0 and make z min 0
    xyz = xyz - xyz.mean(axis=0)
    xyz[:,2] -= xyz[:,2].min()
    # save as .xyz file
    if not xyz_file:
        xyz_file = os.path.join(xyz_dir, os.path.basename(laz_path).split('.')[0] + ".xyz")
    np.savetxt(xyz_file, xyz, fmt="%.6f")
    return xyz_file

def build_helios_scene(config, xyz_files = None, replace = False):
    cnt = 0
    #laz_dir = os.path.join(config["data_dir"], "laz")
    laz_dir = os.path.join(config["scene"]["assets_dir"], "laz")
    xyz_dir = os.path.join(config["scene"]["assets_dir"], "xyz")
    scene_dir = os.path.join(config["simulation"]["data_dir"], "helios_scene")
    os.makedirs(scene_dir, exist_ok=True)
    scene_file = os.path.join(scene_dir, config["scene"]["id"] + ".xml")
    if not replace and os.path.isfile(scene_file):
        return
    tree = ET.parse(os.path.join(sys.path[0], "template/empty_scene.xml"))
    root = tree.getroot()
    scene = list(root)[0]
    scene.attrib["id"] = config["scene"]["id"]
    scene.attrib["name"] = config["scene"]["name"]
    x_offset = config["scene"]["area"]["x_offset"] if "x_offset" in config["scene"]["area"] else 0
    y_offset = config["scene"]["area"]["y_offset"] if "y_offset" in config["scene"]["area"] else 0
    empty_part = list(scene)[0]
    scene.remove(empty_part)
    part_counter = 0
    for tree_set in config["scene"]["trees"]:
        if tree_set["placement"]["method"] == "silva":
            silva_filepath = tree_set["placement"]["filepath"]
            df = pd.read_table(silva_filepath)
            last_period = df['Per'].max()
            df = df.loc[df['Per'] == last_period][['x','y','h','BHD']]
            count = df.shape[0]
        else:            
            count = tree_set["count"]
        for i in range(count):
            if xyz_files:
                laz_file = xyz_files[cnt]
                cnt += 1
            else:
                laz_file = tree_set["path"]
                xyz_file = os.path.join(xyz_dir, '.'.join(laz_file.split('.')[:-1]) + ".xyz")
                if not os.path.isfile(xyz_file):
                    laz_to_xyz(os.path.join(laz_dir,laz_file), xyz_file)
                #laz_file = os.path.join(laz_dir, "%s_%03d.xyz" % (tree_set["template"], i))
            part = deepcopy(empty_part)
            if "part_id" in tree_set:
                part.attrib["id"] = str(tree_set["part_id"])
            else:
                part.attrib["id"] = str(part_counter)
                part_counter += 1
            list(list(part)[0])[0].attrib["value"] = xyz_file
            x,y,z = 0, 0, 0
            scale = 1
            if tree_set["placement"]["method"] == "silva":
                x,y = df.iloc[i][['x','y']]
                z = np.random.normal(0, config["scene"]["area"]["z_var"])
                scale = df.iloc[i]['h']/100.0
            if tree_set["placement"]["method"] == "uniform":
                x,y = np.random.random(2)
                z = np.random.normal(0, config["scene"]["area"]["z_var"])
                x *= config["scene"]["area"]["x"]
                y *= config["scene"]["area"]["y"]
            elif tree_set["placement"]["method"] == "gaussian":
                x,y = np.random.normal(tree_set["placement"]["mean"], tree_set["placement"]["var"], 2)
                z = np.random.normal(0, config["scene"]["area"]["z_var"])
            elif tree_set["placement"]["method"] == "position":
                x = float(tree_set["placement"]["x"]) + x_offset
                y = float(tree_set["placement"]["y"]) + y_offset
                z = np.random.normal(0, config["scene"]["area"]["z_var"])
            list(list(part)[1])[0].attrib["value"] = ";".join([str(x),str(y),str(z)])
            list(list(list(part)[2])[0])[0].attrib["angle_deg"] = str(np.random.random()*360)
            #list(list(part)[3])[0].attrib["value"] = str(scale)
            scene.append(part)
    tree.write(scene_file, xml_declaration=True)

def build_helios_survey(config):
    scene_dir = os.path.join(config["simulation"]["data_dir"], "helios_scene")
    survey_dir = os.path.join(config["simulation"]["data_dir"], "helios_survey")
    os.makedirs(survey_dir, exist_ok=True)
    scene_file = os.path.join(scene_dir, config["scene"]["id"] + ".xml")
    platforms_xml = os.path.join(sys.path[0], "template/platforms.xml")
    scanners_als_xml = os.path.join(sys.path[0], "template/scanners_als.xml")

    tree = ET.parse(os.path.join(sys.path[0], "template/empty_survey.xml"))
    root = tree.getroot()
    platform, scanner, survey = list(root)
    platform.attrib["movePerSec_m"] = str(config["simulation"]["platform"]["speed"])
    platform.attrib["z"] = str(config["simulation"]["platform"]["altitude"])

    scanner.attrib["pulseFreq_hz"] = str(config["simulation"]["scanner"]["pulse_freq"])
    scanner.attrib["scanAngle_deg"] = str(config["simulation"]["scanner"]["angle"])
    #scanner.attrib["scanFreq_hz"] = str(config["survey"]["scanner"]["scan_freq"])
    #scanner.attrib["headRotatePerSec_deg"] = str(config["survey"]["scanner"]["rotate"])

    survey.attrib["name"] = config["simulation"]["name"]
    survey.attrib["scene"] = scene_file + "#" + config["scene"]["id"]
    survey.attrib["platform"] = platforms_xml + "#" + config["simulation"]["platform"]["id"]
    survey.attrib["scanner"] = scanners_als_xml + "#" + config["simulation"]["scanner"]["id"]

    empty_leg = list(survey)[1]
    survey.remove(empty_leg)

    x_offset = config["scene"]["area"]["x_offset"] if "x_offset" in config["scene"]["area"] else 0
    y_offset = config["scene"]["area"]["y_offset"] if "y_offset" in config["scene"]["area"] else 0

    for leg_config in config["simulation"]["legs"]:
        leg = deepcopy(empty_leg)
        leg.attrib["stripId"] = config["simulation"]["name"] + '-full'
        leg_platform = list(leg)[0]
        leg_platform.attrib["x"] = str(float(leg_config["x"] + x_offset))
        leg_platform.attrib["y"] = str(float(leg_config["y"]+ y_offset))
        survey.append(leg)
    survey_file = os.path.join(survey_dir, config["simulation"]["id"] + ".xml")
    tree.write(survey_file, xml_declaration=True)

def run_helios(config):
    helios_exe = os.path.join(config["simulation"]["helios_dir"], "helios")
    survey_dir = os.path.join(config["simulation"]["data_dir"], "helios_survey")
    survey_file = os.path.join(survey_dir, config["simulation"]["id"] + ".xml")
    output_dir = os.path.join(config["simulation"]["data_dir"], "results")
    os.makedirs(output_dir, exist_ok=True)
    cmd = "cd " + config["simulation"]["helios_dir"] + " && " + " ".join(["./helios", survey_file, "--lasOutput", "--chunkSize", "1", "--seed", str(config["simulation"]["helios_seed"]), "--output", output_dir])
    print(cmd)
    os.system(cmd)

def build_scene(config):
    #build_tree_objs(config)
    #xyz_files = get_tree_lazs(config)
    build_helios_scene(config, replace=True)
    build_helios_survey(config)
    run_helios(config)

if __name__ == "__main__":
    if len(sys.argv) < 3:
        help()
        exit()
    scene_file = sys.argv[1]
    simulation_file = sys.argv[2]
    with open(scene_file) as json_file:
        scene = json.load(json_file)
    with open(simulation_file) as json_file:
        simulation = json.load(json_file)
    config = {"scene": scene, "simulation": simulation}
    build_scene(config)
